/*
 *	Armadillo-3x0 GPIO Driver
 *	Copyright (C) 2006 Atmark Techno, Inc.
 */

#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <asm/io.h>

#include <asm/arch/ns9750_bbus.h>
#include <asm/arch/armadillo3x0_gpio.h>

#define AUTHOR      "Atmark Techno, Inc."
#define DESCRIPTION "Armadillo-3x0 GPIO driver"
#define LICENSE     "GPL v2"

#define GPIO_MINOR   (185)
#define GPIO_DEVICE  "gpio"

#define VERSION     DESCRIPTION ", (C) 2005 " AUTHOR

#undef DEBUG
//#define DEBUG
#if defined(DEBUG)
#define DEBUG_FUNC()        printk(GPIO_DEVICE ": %s()\n", __FUNCTION__)
#define DEBUG_INFO(args...) printk(GPIO_DEVICE ": " args)
#else
#define DEBUG_FUNC()
#define DEBUG_INFO(args...)
#endif
#define PRINT_INFO(args...) printk(KERN_INFO GPIO_DEVICE ": " args)
#define PRINT_ERR(args...)  printk(KERN_ERR  GPIO_DEVICE ": " args)

#define INV_NONE   0
#define INV_OUTPUT 1
#define INV_INPUT  2

struct gpio_info {
	int no;
	int gpio;
	int func;
	int inv;
};

static struct gpio_info gpios[] = {
	{ GPIO0, 42, NS_GPIO_CFG_FUNC_GPIO, INV_OUTPUT },
	{ GPIO1, 43, NS_GPIO_CFG_FUNC_GPIO, INV_INPUT  },
	{ GPIO2, 45, NS_GPIO_CFG_FUNC_GPIO, INV_NONE   },
	{ GPIO3, 46, NS_GPIO_CFG_FUNC_GPIO, INV_OUTPUT },
	{ GPIO4, 47, NS_GPIO_CFG_FUNC_GPIO, INV_INPUT  },
	{ GPIO5, 48, NS_GPIO_CFG_FUNC_GPIO, INV_NONE   },
	{ GPIO6, 14, NS_GPIO_CFG_FUNC_GPIO, INV_INPUT  },
	{ GPIO7, 15, NS_GPIO_CFG_FUNC_GPIO, INV_INPUT  },
};

static int 
gpio_open(struct inode *inode, struct file *file)
{
	DEBUG_FUNC();
	return 0;
}

static int 
gpio_release(struct inode *inode, struct file *file)
{
	DEBUG_FUNC();
	return 0;
}

static int 
set_param(struct gpio_param *param){
	int i;
	DEBUG_FUNC();

	switch(param->mode){
	case MODE_OUTPUT:
		for(i=0; i<ARRAY_SIZE(gpios); i++){
			if(param->no == gpios[i].no){
			  set_gpio_cfg(gpios[i].gpio, 
				       gpios[i].func | NS_GPIO_CFG_OUTPUT);

			  set_gpio_ctrl(gpios[i].gpio, 
					(param->data.o.value ? 1 : 0));
			  return 0;
			}
		}
		return -1;
		
	case MODE_INPUT:
		for(i=0; i<ARRAY_SIZE(gpios); i++){
			if(param->no == gpios[i].no){
			  set_gpio_cfg(gpios[i].gpio, 
				       gpios[i].func | NS_GPIO_CFG_INPUT);
			  return 0;
			}
		}
		return -1;
	default:
		return -1;
	}

	return 0;
}

static int 
get_param(struct gpio_param *param){
	int i;
	unsigned long cfg;
	
	DEBUG_FUNC();

	for(i=0; i<ARRAY_SIZE(gpios); i++){
		if(param->no == gpios[i].no){ 
		  cfg = get_gpio_cfg(gpios[i].gpio);
		  
		  if((cfg & NS_GPIO_CFG_FUNC_MA) != NS_GPIO_CFG_FUNC_GPIO){
		    return -1;
		  }

		  switch(cfg & NS_GPIO_CFG_IO_MA){
		  case NS_GPIO_CFG_OUTPUT:
		    param->mode = MODE_OUTPUT;
		    param->data.o.value = get_gpio_ctrl(gpios[i].gpio);
		    break;
		  case NS_GPIO_CFG_INPUT:
		    param->mode = MODE_INPUT;
		    param->data.i.value = get_gpio_stat(gpios[i].gpio);
		    break;
		  }
		  return 0;
		}
	}

	return -EINVAL;
}

static int 
gpio_ioctl(struct inode *inode, struct file *file,
	   unsigned int cmd,unsigned long arg)
{
	struct gpio_param *param_list;
	struct gpio_param *param;
	int ret;
	int i;
	
	DEBUG_FUNC();

	param_list = (struct gpio_param *)arg;
	param = param_list;
	
	switch(cmd){
	case PARAM_SET:
		for(i=0; i<ARRAY_SIZE(gpios); i++){
		  
			if(param == NULL) break;

			ret = set_param(param);
			if(ret) return -EINVAL;
			
			param = param->next;
		}
		break;
		
	case PARAM_GET:
		for(i=0;i<ARRAY_SIZE(gpios);i++){
		  
			if(param == NULL) break;
			
			ret = get_param(param);
			if(ret) return -EINVAL;
			
			param = param->next;
		}
		break;

	default:
		return -EINVAL;
	}
	return 0;
}

static int
a3x0_gpio_request(void)
{
	int i;

	DEBUG_FUNC();

	for(i=0; i<ARRAY_SIZE(gpios); i++){
		if(!request_gpio(gpios[i].gpio, 1, GPIO_DEVICE)){
			goto error;
		}
	}

	for(i=0; i<ARRAY_SIZE(gpios); i++){
		set_gpio_cfg(gpios[i].gpio,
			     gpios[i].func | NS_GPIO_CFG_INPUT);
	}
  
	return 0;

 error:
	while(--i >= 0)
                release_gpio(gpios[i].gpio, 1);
        
        return -EBUSY;
}

static int 
a3x0_gpio_release(void)
{
	int i;

	DEBUG_FUNC();

	for(i=0; i<ARRAY_SIZE(gpios); i++){
		set_gpio_cfg(gpios[i].gpio, 
			     gpios[i].func | NS_GPIO_CFG_INPUT);
		release_gpio(gpios[i].gpio, 1);
	}
	
	return 0;
}

static struct file_operations gpio_fops = {
	.owner   = THIS_MODULE,
	.open    = gpio_open,
	.release = gpio_release,
	.ioctl   = gpio_ioctl,
};

static struct miscdevice gpio_dev = {
	.minor = GPIO_MINOR,
	.name  = GPIO_DEVICE,
	.fops  = &gpio_fops,
};

static int __init 
gpio_init(void)
{
	int ret;

	DEBUG_FUNC();

	ret = a3x0_gpio_request();
	if(ret)
		return -ENODEV;

	if(misc_register(&gpio_dev)){
		return -ENODEV;
	}

	PRINT_INFO(VERSION "\n");
	
	return 0;
}

static void __exit
gpio_exit(void)
{
	DEBUG_FUNC();
	misc_deregister(&gpio_dev);
	a3x0_gpio_release();
}

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);

module_init(gpio_init);
module_exit(gpio_exit);

